import inspect
from dataclasses import dataclass
from typing import Any, get_origin, get_type_hints

from jobify._internal.common.datastructures import RequestState, State
from jobify._internal.configuration import JobifyConfiguration, RouteOptions
from jobify._internal.runners import Runnable
from jobify._internal.scheduler.job import Job


@dataclass(slots=True, kw_only=True)
class JobContext:
    job: Job[Any]
    state: State
    runnable: Runnable[Any]
    request_state: RequestState
    route_options: RouteOptions
    jobify_config: JobifyConfiguration


INJECT: Any = object()
CONTEXT_TYPE_MAP = {
    get_origin(field_type) or field_type: field_name
    for field_name, field_type in get_type_hints(JobContext).items()
}


def inject_context(context: JobContext) -> None:
    bound = context.runnable.bound
    arguments = bound.arguments

    for name, param in bound.signature.parameters.items():
        if param.default is not INJECT:
            continue

        annotation = param.annotation
        if annotation is inspect.Parameter.empty:
            msg = f"Parameter {name} requires a type annotation for INJECT"
            raise ValueError(msg)

        tp = get_origin(annotation) or annotation
        if tp is JobContext:
            val = context
        elif field_name := CONTEXT_TYPE_MAP.get(tp):
            val = getattr(context, field_name)
        else:
            msg = (
                f"Unknown type for injection: {tp}. "
                f"Available types: {list(CONTEXT_TYPE_MAP.keys())}"
            )
            raise ValueError(msg)
        arguments[name] = val



import asyncio
import sqlite3
import threading
from collections.abc import Callable, Sequence
from datetime import datetime
from pathlib import Path
from typing import TYPE_CHECKING, TypeVar
from zoneinfo import ZoneInfo

from typing_extensions import override

from jobify._internal.common.constants import JobStatus
from jobify._internal.storage.abc import (
    ScheduledJob,
    Storage,
    validate_table_name,
)

if TYPE_CHECKING:
    from concurrent.futures import ThreadPoolExecutor

    from jobify._internal.common.types import LoopFactory

CREATE_SCHEDULED_TABLE_QUERY = """
CREATE TABLE IF NOT EXISTS {} (
    job_id TEXT PRIMARY KEY,
    name TEXT,
    message BLOB,
    status TEXT,
    next_run_at TEXT,
    created_at TEXT DEFAULT CURRENT_TIMESTAMP,
    updated_at TEXT DEFAULT CURRENT_TIMESTAMP
);
"""

SELECT_SCHEDULES_QUERY = """
SELECT job_id, name, message, status, next_run_at
FROM {};
"""

INSERT_SCHEDULE_QUERY = """
INSERT INTO {} (job_id, name, message, status, next_run_at)
VALUES (?, ?, ?, ?, ?)
ON CONFLICT (job_id) DO UPDATE SET
    name = EXCLUDED.name,
    message = EXCLUDED.message,
    status = EXCLUDED.status,
    next_run_at = EXCLUDED.next_run_at,
    updated_at = CURRENT_TIMESTAMP;
"""

DELETE_SCHEDULE_QUERY = """
DELETE FROM {} WHERE job_id = ?;
"""

DELETE_SCHEDULE_MANY_QUERY = """
DELETE FROM {table_name} WHERE job_id IN ({placeholder});
"""

ReturnT = TypeVar("ReturnT")


class SQLiteStorage(Storage):
    def __init__(
        self,
        database: str | Path = "jobify.db",
        *,
        table_name: str = "jobify_schedules",
        timeout: float = 20.0,
    ) -> None:
        validate_table_name(table_name)
        self.database: Path = (
            Path(database) if isinstance(database, str) else database
        )
        self.table_name: str = table_name
        self.timeout: float = timeout
        self.tz: ZoneInfo = ZoneInfo("UTC")
        self.getloop: LoopFactory = asyncio._get_running_loop
        self.threadpool: ThreadPoolExecutor | None = None
        self._conn: sqlite3.Connection | None = None
        self._lock: threading.Lock = threading.Lock()

        self.create_scheduled_table_query: str = (
            CREATE_SCHEDULED_TABLE_QUERY.format(table_name)
        )
        self.select_schedules_query: str = SELECT_SCHEDULES_QUERY.format(
            table_name,
        )
        self.insert_schedule_query: str = INSERT_SCHEDULE_QUERY.format(
            table_name,
        )
        self.delete_schedule_query: str = DELETE_SCHEDULE_QUERY.format(
            table_name,
        )

    @property
    def conn(self) -> sqlite3.Connection:
        if self._conn is None:
            msg = "Database not initialized. Call startup() first."
            raise RuntimeError(msg)
        return self._conn

    async def _to_thread(self, func: Callable[[], ReturnT]) -> ReturnT:
        def thread_safe() -> ReturnT:
            with self._lock:
                return func()

        loop = self.getloop()
        return await loop.run_in_executor(self.threadpool, thread_safe)

    @override
    async def startup(self) -> None:
        conn = sqlite3.connect(
            database=self.database,
            timeout=self.timeout,
            check_same_thread=False,
        )
        _ = conn.execute("PRAGMA journal_mode=WAL;")
        _ = conn.execute("PRAGMA synchronous=NORMAL;")
        _ = conn.execute(self.create_scheduled_table_query)
        conn.commit()
        self._conn = conn

    @override
    async def shutdown(self) -> None:
        if self._conn is not None:
            with self._lock:
                self._conn.close()
                self._conn = None

    @override
    async def get_schedules(self) -> list[ScheduledJob]:
        def get() -> list[ScheduledJob]:
            cursor = self.conn.execute(self.select_schedules_query)
            return [
                ScheduledJob(
                    job_id=row[0],
                    name=row[1],
                    message=row[2],
                    status=JobStatus(row[3]),
                    next_run_at=(
                        datetime.fromisoformat(row[4]).astimezone(self.tz)
                    ),
                )
                for row in cursor.fetchall()
            ]

        return await self._to_thread(get)

    @override
    async def add_schedule(self, *scheduled: ScheduledJob) -> None:
        def insert_many() -> None:
            with self.conn as conn:
                _ = conn.executemany(
                    self.insert_schedule_query,
                    [
                        (
                            sch.job_id,
                            sch.name,
                            sch.message,
                            sch.status,
                            sch.next_run_at.isoformat(),
                        )
                        for sch in scheduled
                    ],
                )

        return await self._to_thread(insert_many)

    @override
    async def delete_schedule(self, job_id: str) -> None:
        def delete() -> None:
            with self.conn as conn:
                _ = conn.execute(self.delete_schedule_query, (job_id,))

        return await self._to_thread(delete)

    @override
    async def delete_schedule_many(self, job_ids: Sequence[str]) -> None:
        if not job_ids:
            return None

        job_ids = sorted(job_ids)

        def delete_many() -> None:
            batch_size = 500
            with self.conn as conn:
                for i in range(0, len(job_ids), batch_size):
                    batch = job_ids[i : i + batch_size]
                    query = DELETE_SCHEDULE_MANY_QUERY.format_map(
                        {
                            "table_name": self.table_name,
                            "placeholder": ",".join("?" * len(batch)),
                        }
                    )
                    _ = conn.execute(query, batch)

        return await self._to_thread(delete_many)



"""Jobify entrypoint."""

from __future__ import annotations

import asyncio
import functools
import logging
import signal
import sys
import threading
from contextlib import contextmanager
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Literal, ParamSpec, TypeVar
from zoneinfo import ZoneInfo

from typing_extensions import Self

from jobify._internal.common.constants import EMPTY, PATCH_CRON_DEF_ID
from jobify._internal.configuration import (
    Cron,
    JobifyConfiguration,
    WorkerPools,
)
from jobify._internal.message import AtArguments, CronArguments, Message
from jobify._internal.router.root import (
    CRONS_DEF_KEY,
    CronsDefinition,
    RootRouter,
)
from jobify._internal.scheduler.misfire_policy import (
    GracePolicy,
    MisfirePolicy,
    handle_misfire_policy,
)
from jobify._internal.serializers.json import JSONSerializer
from jobify._internal.serializers.json_extended import ExtendedJSONSerializer
from jobify._internal.shared_state import SharedState
from jobify._internal.storage.abc import ScheduledJob
from jobify._internal.storage.dummy import DummyStorage
from jobify._internal.storage.sqlite import SQLiteStorage
from jobify._internal.typeadapter.dummy import DummyDumper, DummyLoader
from jobify.crontab import create_crontab

if TYPE_CHECKING:
    from collections.abc import Callable, Iterator, Sequence
    from concurrent.futures import ProcessPoolExecutor, ThreadPoolExecutor
    from datetime import datetime
    from types import FrameType, TracebackType

    from jobify._internal.common.types import Lifespan, LoopFactory
    from jobify._internal.cron_parser import CronFactory, CronParser
    from jobify._internal.middleware.base import BaseMiddleware
    from jobify._internal.middleware.exceptions import MappingExceptionHandlers
    from jobify._internal.scheduler.job import Job
    from jobify._internal.scheduler.scheduler import ScheduleBuilder
    from jobify._internal.serializers.base import Serializer
    from jobify._internal.storage.abc import Storage
    from jobify._internal.typeadapter.base import Dumper, Loader

HANDLED_SIGNALS = (
    signal.SIGINT,  # Unix signal 2. Sent by Ctrl+C.
    signal.SIGTERM,  # Unix signal 15. Sent by `kill <pid>`.
)
if sys.platform == "win32":  # pragma: no cover
    # Windows signal 21. Sent by Ctrl+Break.
    HANDLED_SIGNALS += (signal.SIGBREAK,)  # pyright: ignore[reportConstantRedefinition]

logger = logging.getLogger("Jobify")

AppT = TypeVar("AppT", bound="Jobify")
ReturnT = TypeVar("ReturnT")
ParamsT = ParamSpec("ParamsT")


@dataclass(slots=True)
class _CronSchedule:
    arg: CronArguments
    builder: ScheduleBuilder[Any]
    next_run_at: datetime
    cron_parser: CronParser


def cache_result(f: Callable[ParamsT, ReturnT]) -> Callable[ParamsT, ReturnT]:
    """Cache the result of the first function call."""
    result: ReturnT | None = None

    @functools.wraps(f)
    def wrapper(*args: ParamsT.args, **kwargs: ParamsT.kwargs) -> ReturnT:
        nonlocal result
        if result is None:
            result = f(*args, **kwargs)
        return result

    return wrapper


class Jobify(RootRouter):
    """Jobify is the main app for scheduling and managing background jobs.

    It provides a flexible and extensible framework for defining, running,
    and persisting jobs, supporting various executors, middleware, and
    serialization options.
    """

    def __init__(  # noqa: PLR0913
        self,
        *,
        tz: ZoneInfo | None = None,
        dumper: Dumper | None = None,
        loader: Loader | None = None,
        storage: Storage | Literal[False] = EMPTY,
        lifespan: Lifespan[AppT] | None = None,
        serializer: Serializer | None = None,
        middleware: Sequence[BaseMiddleware] | None = None,
        cron_factory: CronFactory = create_crontab,
        loop_factory: LoopFactory = asyncio.get_running_loop,
        exception_handlers: MappingExceptionHandlers | None = None,
        threadpool_executor: ThreadPoolExecutor | None = None,
        processpool_executor: ProcessPoolExecutor | None = None,
    ) -> None:
        """Initialize a `Jobify` instance."""
        getloop = cache_result(loop_factory)
        tz = tz or ZoneInfo("UTC")

        if storage is False:
            storage = DummyStorage()
        elif storage is EMPTY:
            storage = SQLiteStorage()

        if isinstance(storage, SQLiteStorage):
            storage.getloop = getloop
            storage.threadpool = threadpool_executor
            storage.tz = tz

        if serializer is None:
            system_types = (
                Message,
                Cron,
                CronArguments,
                AtArguments,
                MisfirePolicy,
                GracePolicy,
            )
            serializer = (
                ExtendedJSONSerializer(system_types)
                if dumper is None and loader is None
                else JSONSerializer()
            )

        if dumper is None:
            dumper = DummyDumper()
        if loader is None:
            loader = DummyLoader()

        self.configs: JobifyConfiguration = JobifyConfiguration(
            tz=tz,
            dumper=dumper,
            loader=loader,
            storage=storage,
            getloop=getloop,
            serializer=serializer,
            worker_pools=WorkerPools(
                _processpool=processpool_executor,
                threadpool=threadpool_executor,
            ),
            cron_factory=cron_factory,
        )
        super().__init__(
            lifespan=lifespan,
            middleware=middleware,
            shared_state=SharedState(),
            jobify_config=self.configs,
            exception_handlers=exception_handlers,
        )
        self._captured_signals: list[int] = []

    def find_job(self, id_: str, /) -> Job[ReturnT] | None:
        """Find an active job by its ID.

        Args:
            id_: Unique identifier of the job.

        Returns:
            The `Job` instance if it's currently pending or running,
            otherwise `None`.

        """
        return self.task._shared_state.pending_jobs.get(id_)

    def get_active_jobs(self) -> list[Job[Any]]:
        """Return a list of all currently active jobs."""
        return list(self.task._shared_state.pending_jobs.values())

    async def __aenter__(self) -> Self:
        """Enter the Jobify context manager.

        Returns:
            The initialized Jobify instance ready for use.

        Raises:
            Any exception raised by `startup()` method.

        """
        await self.startup()
        return self

    async def startup(self) -> None:
        """Initialize the Jobify application.

        This method:
        1. Marks the application as started
        2. Propagates startup events to all routers and their registrators
        3. Schedules any pending cron jobs

        Raises:
            RuntimeError: If application startup fails due to configuration
            issues or router initialization errors.

        """
        self.configs.app_started = True
        await self.configs.storage.startup()
        await self._propagate_startup(self)
        await self._restore_schedules()

    async def _restore_schedules(self) -> None:
        crons_def: CronsDefinition = self.state.pop(CRONS_DEF_KEY, {})
        schedules = await self.configs.storage.get_schedules()
        db_map = {sch.job_id: sch for sch in schedules}

        at_args: dict[str, tuple[ScheduleBuilder[Any], AtArguments]] = {}
        crons_args: dict[str, _CronSchedule] = {}
        crons_def_persist: set[str] = set()
        scheduled_to_update: list[ScheduledJob] = []
        scheduled_to_delete: list[str] = []

        for job_id, (rout, cron) in crons_def.items():
            builder = rout.schedule()
            offset = builder.now()

            cron_parser = self.configs.cron_factory(cron.expression)
            next_run_at = cron_parser.next_run(now=offset)
            trigger = CronArguments(cron, job_id, offset)
            crons_args[job_id] = _CronSchedule(
                trigger,
                builder,
                next_run_at,
                cron_parser,
            )
            if builder._is_persist():
                if job_id in db_map:
                    crons_def_persist.add(job_id)
                else:
                    scheduled_to_update.append(
                        ScheduledJob.create(
                            job_id,
                            builder.name,
                            builder._serialize_job_message(trigger),
                            next_run_at,
                        )
                    )

        loadb = self.configs.serializer.loadb
        type_adaptive = self.configs.loader.load

        for job_id, sch in db_map.items():
            route = self.task._routes.get(sch.name, ...)
            if (
                route is ...
                or route.options.get("durable") is False
                or (
                    job_id.endswith(PATCH_CRON_DEF_ID)
                    and job_id not in crons_def
                )
            ):
                scheduled_to_delete.append(job_id)
                continue
            try:
                msg = type_adaptive(loadb(sch.message), Message)
                for name, raw_arg in msg.arguments.items():
                    param_type = route.func_spec.params_type[name]
                    msg.arguments[name] = type_adaptive(raw_arg, param_type)
                bound = route.func_spec.signature.bind(**msg.arguments)
            except (KeyError, TypeError, ValueError) as exc:
                # KeyError: The function has been removed from the router
                #   (the code has changed).
                # TypeError: The arguments in the database do not match the new
                #   function signature.
                # ValueError: Serializer error.
                warn = (
                    f"Cannot restore <job_id {job_id!r}>"
                    f"<name {sch.name!r}>.\n"
                    f"Exception Type: {type(exc)}. "
                    f"Reason: {exc}. Removing from storage."
                )
                logger.warning(warn)
                scheduled_to_delete.append(job_id)
            except Exception:  # pragma: no cover
                err = f"Unexpected error restoring job {job_id}"
                logger.exception(err)
                scheduled_to_delete.append(job_id)
            else:
                builder = route.create_builder(bound)
                self._handle_trigger(
                    msg.trigger,
                    builder,
                    scheduled_to_update,
                    crons_def_persist,
                    crons_args,
                    at_args,
                )
        await self.configs.storage.add_schedule(*scheduled_to_update)
        await self.configs.storage.delete_schedule_many(scheduled_to_delete)
        self._start_pending_schedules(crons_args, at_args)

    def _handle_trigger(  # noqa: PLR0913
        self,
        trigger: CronArguments | AtArguments,
        builder: ScheduleBuilder[Any],
        scheduled_to_update: list[ScheduledJob],
        crons_declare_persist: set[str],
        crons_args: dict[str, _CronSchedule],
        at_args: dict[str, tuple[ScheduleBuilder[Any], AtArguments]],
    ) -> None:
        match trigger:
            case CronArguments(db_cron, job_id, db_offset) as trigger:
                if job_id in crons_declare_persist:
                    origin = crons_args[job_id]
                    new_cron = origin.arg.cron
                    new_builder = origin.builder

                    offset, next_run_at = new_builder._calculate_next_run_at(
                        old_cron=db_cron,
                        new_cron=new_cron,
                        real_now=new_builder.now(),
                        parser=origin.cron_parser,
                        offset=db_offset,
                    )
                    trigger.cron = new_cron
                    trigger.run_count = 0
                    trigger.offset = offset
                    origin.arg.run_count = 0
                    origin.arg.offset = offset
                    origin.next_run_at = next_run_at
                    scheduled_to_update.append(
                        ScheduledJob.create(
                            job_id,
                            builder.name,
                            builder._serialize_job_message(trigger),
                            next_run_at,
                        )
                    )
                else:
                    parser = self.configs.cron_factory(db_cron.expression)
                    next_run_at = handle_misfire_policy(
                        parser,
                        parser.next_run(now=db_offset),
                        builder.now(),
                        db_cron.misfire_policy,
                    )
                    crons_args[job_id] = _CronSchedule(
                        trigger,
                        builder,
                        next_run_at,
                        parser,
                    )
            case AtArguments(_, job_id) as trigger:
                at_args[job_id] = (builder, trigger)

    def _start_pending_schedules(
        self,
        crons_args: dict[str, _CronSchedule],
        at_args: dict[str, tuple[ScheduleBuilder[Any], AtArguments]],
    ) -> None:
        for cron in crons_args.values():
            _ = cron.builder._create_cron_job(
                cron=cron.arg.cron,
                job_id=cron.arg.job_id,
                next_run_at=cron.next_run_at,
                cron_parser=cron.cron_parser,
                offset=cron.arg.offset,
                run_count=cron.arg.run_count,
            )
        for builder, arg in at_args.values():
            _ = builder._at(arg.at, arg.job_id)

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None = None,
        exc_val: BaseException | None = None,
        exc_tb: TracebackType | None = None,
    ) -> None:
        """Exit the Jobify context manager.

        Note:
            This method ensures proper shutdown regardless of whether an
            exception occurred in the managed context. The exception parameters
            are ignored as shutdown should proceed even if the context failed.

        """
        await self.shutdown()

    async def shutdown(self) -> None:
        """Gracefully shut down the Jobify application.

        This method performs a structured shutdown:
        1. Marks the application as stopped (`app_started = False`).
        2. Propagates shutdown events to all routers/components.
        3. Cancels all scheduled future jobs in the registry
           (`_jobs_registry`).
        4. Closes the jobify configuration (e.g., stopping the internal
           scheduler).
        5. Cancels all currently running tasks (in `_tasks_registry`), waits
           for their completion, and explicitly clears the task registry.

        Note:
            The method uses `return_exceptions=True` when gathering cancelled
            tasks to prevent shutdown from being interrupted by task exception.

        """
        self.configs.app_started = False

        if jobs := tuple(self.task._shared_state.pending_jobs.values()):
            for job in jobs:
                job._cancel()

        if tasks := self.task._shared_state.pending_tasks:  # pragma: no cover
            for task in tuple(tasks):
                _ = task.cancel()
            _ = await asyncio.gather(*tasks, return_exceptions=True)

        self.configs.worker_pools.close()
        await self._propagate_shutdown()
        await self.configs.storage.shutdown()

        # If we did gracefully shut down due to a signal, try to
        # trigger the expected behaviour now; multiple signals would be
        # done LIFO, see https://stackoverflow.com/questions/48434964
        for captured_signal in reversed(self._captured_signals):
            signal.raise_signal(captured_signal)

    async def wait_all(self, timeout: float | None = None) -> None:
        """Wait for all currently scheduled jobs to complete.

        This method waits until all jobs currently registered have finished
        executing (with statuses of SUCCESS, FAILED, or TIMEOUT). This is
        useful in situations where it's important to ensure that background
        tasks have completed before moving on.

        The method sets an internal event when both conditions are met:
        1. No jobs remain in the jobs registry (`_jobs_registry`)

        Args:
            timeout (optional): The maximum time in seconds to wait for the
                jobs to complete. If not specified, the default value of `None`
                will be used, which means the job will wait indefinitely. If a
                timeout is specified and it is reached, the method will raise
                an `asyncio.TimeoutError`.

        """
        idle_event = self.task._shared_state.idle_event
        with self._capture_signals():
            _ = await asyncio.wait_for(idle_event.wait(), timeout=timeout)

    @contextmanager
    def _capture_signals(self) -> Iterator[None]:
        # Signals can only be listened to from the main thread.
        if threading.current_thread() is not threading.main_thread():
            yield
            return
        # always use signal.signal, even if loop.add_signal_handler is
        # available this allows to restore previous signal handlers later on
        original_handlers = {
            sig: signal.signal(sig, self._handle_exit)
            for sig in HANDLED_SIGNALS
        }
        try:
            yield
        finally:
            for sig, handler in original_handlers.items():
                _ = signal.signal(sig, handler)

    def _handle_exit(self, sig: int, _: FrameType | None) -> None:
        self._captured_signals.append(sig)
        loop = self.configs.getloop()
        idle_event = self.task._shared_state.idle_event
        _handle = loop.call_soon_threadsafe(idle_event.set)



from __future__ import annotations

import asyncio
from collections.abc import Awaitable, Callable, Mapping
from typing import TYPE_CHECKING, Any, TypeAlias, final

from typing_extensions import override

from jobify._internal.context import JobContext
from jobify._internal.middleware.base import BaseMiddleware, CallNext

if TYPE_CHECKING:
    from jobify._internal.configuration import JobifyConfiguration


ExceptionHandler: TypeAlias = Callable[
    [Exception, JobContext], Awaitable[None] | None
]
ExceptionHandlers: TypeAlias = dict[type[Exception], ExceptionHandler]
MappingExceptionHandlers: TypeAlias = Mapping[
    type[Exception], ExceptionHandler
]


@final
class ExceptionMiddleware(BaseMiddleware):
    __slots__: tuple[str, ...] = ("exc_handlers", "jobify_config")

    def __init__(
        self,
        exc_handlers: ExceptionHandlers,
        jobify_config: JobifyConfiguration,
    ) -> None:
        self.exc_handlers = exc_handlers
        self.jobify_config = jobify_config

    @override
    async def __call__(self, call_next: CallNext, context: JobContext) -> Any:
        try:
            return await call_next(context)
        except Exception as exc:
            handler = self._lookup_exc_handler(exc)
            if handler:
                if asyncio.iscoroutinefunction(handler):
                    await handler(exc, context)
                else:
                    loop = self.jobify_config.getloop()
                    thread = self.jobify_config.worker_pools.threadpool
                    await loop.run_in_executor(thread, handler, exc, context)  # pyright: ignore[reportUnusedCallResult]
            raise

    def _lookup_exc_handler(self, exc: Exception) -> ExceptionHandler | None:
        for cls_exc in type(exc).__mro__:
            if handler := self.exc_handlers.get(cls_exc):
                return handler
        return None



import inspect
from collections.abc import Callable
from dataclasses import dataclass
from typing import Any, Generic, ParamSpec, TypeAlias, TypeVar, get_type_hints

ReturnT = TypeVar("ReturnT")
ParamsT = ParamSpec("ParamsT")

ParamName: TypeAlias = str
TypeHint: TypeAlias = Any


@dataclass(slots=True, kw_only=True)
class FuncSpec(Generic[ReturnT]):
    name: str
    signature: inspect.Signature
    params_type: dict[ParamName, TypeHint]
    result_type: type[ReturnT]


def get_params_type(
    sig: inspect.Signature,
    hints: dict[str, Any],
) -> dict[ParamName, TypeHint]:
    return {
        arg.name: hints.get(arg.name, Any) for arg in sig.parameters.values()
    }


def get_result_type(hints: dict[str, Any]) -> Any:  # noqa: ANN401
    return hints.get("return", Any)


def make_func_spec(func: Callable[ParamsT, ReturnT]) -> FuncSpec[ReturnT]:
    sig = inspect.signature(func)
    hints = get_type_hints(func)
    return FuncSpec(
        name=func.__name__,
        params_type=get_params_type(sig, hints),
        result_type=get_result_type(hints),
        signature=sig,
    )



from __future__ import annotations

import asyncio
import logging
from datetime import datetime, timedelta
from typing import TYPE_CHECKING, Generic, TypeVar, cast
from uuid import uuid4

from jobify._internal.common.constants import JobStatus
from jobify._internal.common.datastructures import RequestState, State
from jobify._internal.configuration import Cron
from jobify._internal.context import JobContext
from jobify._internal.exceptions import DuplicateJobError, JobTimeoutError
from jobify._internal.message import AtArguments, CronArguments, Message
from jobify._internal.scheduler.job import CronContext, Job
from jobify._internal.scheduler.misfire_policy import handle_misfire_policy
from jobify._internal.storage.abc import ScheduledJob
from jobify._internal.storage.dummy import DummyStorage

if TYPE_CHECKING:
    from jobify._internal.configuration import (
        JobifyConfiguration,
        RouteOptions,
    )
    from jobify._internal.cron_parser import CronParser
    from jobify._internal.inspection import FuncSpec
    from jobify._internal.middleware.base import CallNext
    from jobify._internal.runners import Runnable
    from jobify._internal.serializers.base import JSONCompat
    from jobify._internal.shared_state import SharedState


ReturnT = TypeVar("ReturnT")
logger = logging.getLogger("jobify.scheduler")


class ScheduleBuilder(Generic[ReturnT]):
    __slots__: tuple[str, ...] = (
        "_chain_middleware",
        "_configs",
        "_runnable",
        "_shared_state",
        "_state",
        "func_spec",
        "name",
        "route_options",
    )

    def __init__(  # noqa: PLR0913
        self,
        *,
        state: State,
        shared_state: SharedState,
        jobify_config: JobifyConfiguration,
        runnable: Runnable[ReturnT],
        chain_middleware: CallNext,
        func_spec: FuncSpec[ReturnT],
        options: RouteOptions,
        name: str,
    ) -> None:
        self._state: State = state
        self._shared_state: SharedState = shared_state
        self._configs: JobifyConfiguration = jobify_config
        self._runnable: Runnable[ReturnT] = runnable
        self._chain_middleware: CallNext = chain_middleware
        self.func_spec: FuncSpec[ReturnT] = func_spec
        self.route_options: RouteOptions = options
        self.name: str = name

    def now(self) -> datetime:
        return datetime.now(tz=self._configs.tz)

    def _calculate_delay_seconds(self, target_at: datetime) -> float:
        return target_at.timestamp() - self.now().timestamp()

    def _is_persist(self) -> bool:
        is_dummy = isinstance(self._configs.storage, DummyStorage)
        is_durable = self.route_options.get("durable", True)
        return not is_dummy and is_durable

    async def _persist_job(
        self,
        job_id: str,
        next_run_at: datetime,
        trigger: CronArguments | AtArguments,
    ) -> None:
        await self._configs.storage.add_schedule(
            ScheduledJob.create(
                job_id,
                self.name,
                self._serialize_job_message(trigger),
                next_run_at,
            )
        )

    def _serialize_job_message(
        self,
        trigger: CronArguments | AtArguments,
    ) -> bytes:
        raw_args = self._runnable.origin_arguments
        params_type = self.func_spec.params_type
        dumper_hook = self._configs.dumper.dump
        dumped_args: dict[str, JSONCompat] = {
            name: dumper_hook(arg_val, params_type[name])
            for name, arg_val in raw_args.items()
        }
        msg = Message(
            job_id=trigger.job_id,
            name=self.name,
            arguments=dumped_args,
            trigger=trigger,
        )
        formatted_msg = self._configs.dumper.dump(msg, Message)
        return self._configs.serializer.dumpb(formatted_msg)

    def _ensure_job_id(
        self,
        job_id: str | None,
        *,
        replace: bool,
    ) -> tuple[str, Job[ReturnT] | None]:
        job_id = job_id or uuid4().hex
        if job := self._shared_state.pending_jobs.get(job_id):
            if replace is True:
                if job._handle is not None:
                    job._handle.cancel()
                return (job_id, job)

            raise DuplicateJobError(job_id)

        return (job_id, None)

    async def cron(
        self,
        cron: str | Cron,
        *,
        job_id: str,
        replace: bool = False,
    ) -> Job[ReturnT]:
        job_id, exists_job = self._ensure_job_id(job_id, replace=replace)
        if isinstance(cron, str):
            cron = Cron(cron)
        parser = self._configs.cron_factory(cron.expression)
        real_now = self.now()

        if exists_job is not None:
            return await self._update_exists_cron_job(
                exists_job,
                cron,
                parser,
                real_now,
            )

        if cron.start_date is not None:
            target_time = cron.start_date
            offset = target_time
        else:
            offset = real_now
            target_time = parser.next_run(now=offset)

        next_run_at = handle_misfire_policy(
            parser,
            target_time,
            real_now,
            cron.misfire_policy,
        )
        if self._is_persist():
            trigger = CronArguments(cron=cron, job_id=job_id, offset=offset)
            await self._persist_job(job_id, next_run_at, trigger)

        return self._create_cron_job(
            cron=cron,
            job_id=job_id,
            offset=offset,
            next_run_at=next_run_at,
            cron_parser=parser,
            run_count=0,
        )

    async def _update_exists_cron_job(
        self,
        job: Job[ReturnT],
        new_cron: Cron,
        parser: CronParser,
        real_now: datetime,
    ) -> Job[ReturnT]:
        ctx = cast("CronContext[ReturnT]", job._cron_context)
        offset, next_run_at = self._calculate_next_run_at(
            old_cron=ctx.cron,
            new_cron=new_cron,
            real_now=real_now,
            parser=parser,
            offset=ctx.offset,
        )
        job.exec_at = next_run_at
        ctx.offset = offset
        ctx.run_count = 0
        ctx.cron = new_cron
        ctx.cron_parser = parser

        if self._is_persist():
            trigger = CronArguments(
                cron=new_cron,
                job_id=job.id,
                offset=ctx.offset,
                run_count=0,
            )
            await self._persist_job(job.id, next_run_at, trigger)

        self._schedule_execution_cron(ctx)
        return job

    def _calculate_next_run_at(
        self,
        old_cron: Cron,
        new_cron: Cron,
        real_now: datetime,
        parser: CronParser,
        offset: datetime,
    ) -> tuple[datetime, datetime]:
        start_date_changed = new_cron.start_date != old_cron.start_date
        if start_date_changed and new_cron.start_date is not None:
            target_at = new_cron.start_date
            offset = target_at
        else:
            target_at = parser.next_run(now=offset)

        next_run_at = handle_misfire_policy(
            parser,
            target_at,
            real_now,
            new_cron.misfire_policy,
        )
        return (offset, next_run_at)

    def _create_cron_job(  # noqa: PLR0913
        self,
        *,
        cron: Cron,
        job_id: str,
        offset: datetime,
        next_run_at: datetime,
        cron_parser: CronParser,
        run_count: int,
    ) -> Job[ReturnT]:
        job = Job[ReturnT](
            exec_at=next_run_at,
            job_id=job_id,
            unregister_hook=self._shared_state.unregister_job,
            storage=self._configs.storage,
        )
        cron_ctx = CronContext(
            job=job,
            cron=cron,
            offset=offset,
            cron_parser=cron_parser,
            run_count=run_count,
        )
        job.bind_cron_context(cron_ctx)
        self._shared_state.register_job(job)
        self._schedule_execution_cron(cron_ctx)
        return job

    def _schedule_execution_cron(self, ctx: CronContext[ReturnT]) -> None:
        delay_seconds = self._calculate_delay_seconds(ctx.job.exec_at)
        loop = self._configs.getloop()
        when = loop.time() + delay_seconds
        handle = loop.call_at(when, self._pre_exec_cron, ctx)
        ctx.job.bind_handle(handle)

    async def delay(
        self,
        seconds: float,
        *,
        job_id: str | None = None,
        now: datetime | None = None,
        replace: bool = False,
    ) -> Job[ReturnT]:
        now = now or self.now()
        at = now + timedelta(seconds=seconds)
        return await self.at(at=at, job_id=job_id, replace=replace)

    async def at(
        self,
        at: datetime,
        *,
        job_id: str | None = None,
        replace: bool = False,
    ) -> Job[ReturnT]:
        job_id, exists_job = self._ensure_job_id(job_id, replace=replace)

        if self._is_persist():
            trigger = AtArguments(at=at, job_id=job_id)
            await self._persist_job(job_id, at, trigger)

        if exists_job is not None:
            exists_job.exec_at = at
            self._schedule_execution_at(exists_job)
            return exists_job

        job = Job[ReturnT](
            exec_at=at,
            job_id=job_id,
            unregister_hook=self._shared_state.unregister_job,
            storage=self._configs.storage,
        )
        self._schedule_execution_at(job)
        self._shared_state.register_job(job)
        return job

    def _at(self, at: datetime, job_id: str) -> Job[ReturnT]:
        job = Job[ReturnT](
            exec_at=at,
            job_id=job_id,
            unregister_hook=self._shared_state.unregister_job,
            storage=self._configs.storage,
        )
        self._schedule_execution_at(job)
        self._shared_state.register_job(job)
        return job

    def _schedule_execution_at(self, job: Job[ReturnT]) -> None:
        loop = self._configs.getloop()
        delay_seconds = self._calculate_delay_seconds(target_at=job.exec_at)
        if delay_seconds <= 0:
            handle = loop.call_soon(self._pre_exec_at, job)
        else:
            when = loop.time() + delay_seconds
            handle = loop.call_at(when, self._pre_exec_at, job)
        job.bind_handle(handle)

    def _pre_exec_at(self, job: Job[ReturnT]) -> None:
        task = asyncio.create_task(self._exec_at(job), name=job.id)
        self._shared_state.track_task(task, job._event)

    async def _exec_at(self, job: Job[ReturnT]) -> None:
        await self._exec_job(job)
        self._shared_state.unregister_job(job.id)
        if self._is_persist():
            await self._configs.storage.delete_schedule(job.id)

    def _pre_exec_cron(self, ctx: CronContext[ReturnT]) -> None:
        task = asyncio.create_task(self._exec_cron(ctx=ctx), name=ctx.job.id)
        self._shared_state.track_task(task, ctx.job._event)

    async def _exec_cron(self, ctx: CronContext[ReturnT]) -> None:
        job = ctx.job
        await self._exec_job(job)

        if job.status is JobStatus.SUCCESS:
            ctx.failure_count = 0
        else:
            ctx.failure_count += 1
        ctx.run_count += 1

        if ctx.is_run_exceeded_by_limit():
            if self._is_persist():
                await self._configs.storage.delete_schedule(job.id)
        elif job.is_reschedulable() and self._configs.app_started:
            if ctx.is_failure_allowed_by_limit():
                offset = ctx.offset = job.exec_at
                next_run_at = ctx.cron_parser.next_run(now=offset)
                if self._is_persist():
                    trigger = CronArguments(
                        ctx.cron,
                        job.id,
                        offset,
                        ctx.run_count,
                    )
                    await self._persist_job(job.id, next_run_at, trigger)
                self._reschedule_cron(ctx, next_run_at)
                return

            job._status = JobStatus.PERMANENTLY_FAILED
            logger.warning(
                "Job %s stopped due to max failures policy (%s/%s)",
                job.id,
                ctx.failure_count,
                ctx.cron.max_failures,
            )
        self._shared_state.unregister_job(job.id)

    def _reschedule_cron(
        self,
        ctx: CronContext[ReturnT],
        next_run_at: datetime,
    ) -> None:
        delay_seconds = self._calculate_delay_seconds(target_at=next_run_at)
        loop = self._configs.getloop()
        when = loop.time() + delay_seconds
        handle = loop.call_at(when, self._pre_exec_cron, ctx)
        job = ctx.job
        job.update(
            exec_at=next_run_at,
            time_handler=handle,
            job_status=JobStatus.SCHEDULED,
        )

    async def _exec_job(self, job: Job[ReturnT]) -> None:
        job._status = JobStatus.RUNNING
        job_context = JobContext(
            job=job,
            state=self._state,
            request_state=RequestState(),
            runnable=self._runnable,
            route_options=self.route_options,
            jobify_config=self._configs,
        )
        try:
            result = await self._chain_middleware(job_context)
        except JobTimeoutError as exc:
            job.set_exception(exc, status=JobStatus.TIMEOUT)
        except Exception as exc:
            logger.exception("Job %s failed with unexpected error", job.id)
            job.set_exception(exc, status=JobStatus.FAILED)
        else:
            job.set_result(result, status=JobStatus.SUCCESS)



from __future__ import annotations

import functools
from abc import ABCMeta, abstractmethod
from collections.abc import Awaitable, Callable, Sequence
from typing import Any, Protocol, TypeAlias, TypeVar, runtime_checkable

from jobify._internal.context import JobContext

ReturnT = TypeVar("ReturnT")
CallNext: TypeAlias = Callable[[JobContext], Awaitable[Any]]


@runtime_checkable
class BaseMiddleware(Protocol, metaclass=ABCMeta):
    @abstractmethod
    async def __call__(self, call_next: CallNext, context: JobContext) -> Any:  # noqa: ANN401
        pass


def build_middleware(
    middleware: Sequence[BaseMiddleware],
    /,
    func: Callable[[JobContext], Awaitable[ReturnT]],
) -> CallNext:
    chain_of_middlewares = func
    for m in reversed(middleware):
        chain_of_middlewares = functools.partial(m, chain_of_middlewares)

    return chain_of_middlewares



from dataclasses import dataclass
from datetime import datetime
from typing import Any, NamedTuple

from jobify._internal.configuration import Cron


@dataclass(slots=True)
class CronArguments:
    cron: Cron
    job_id: str
    offset: datetime
    run_count: int = 0


class AtArguments(NamedTuple):
    at: datetime
    job_id: str


@dataclass(slots=True, kw_only=True)
class Message:
    job_id: str
    name: str
    arguments: dict[str, Any]
    trigger: CronArguments | AtArguments



from __future__ import annotations

import asyncio
import functools
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from jobify._internal.scheduler.job import Job


@dataclass(slots=True, kw_only=True, frozen=True)
class SharedState:
    pending_jobs: dict[str, Job[Any]] = field(default_factory=dict)
    pending_tasks: set[asyncio.Task[Any]] = field(default_factory=set)
    idle_event: asyncio.Event = field(default_factory=asyncio.Event)

    def __post_init__(self) -> None:
        self.idle_event.set()

    def register_job(self, job: Job[Any]) -> None:
        self.idle_event.clear()
        self.pending_jobs[job.id] = job

    def unregister_job(self, job_id: str) -> None:
        if self.pending_jobs.pop(job_id, None):
            self._check_state()

    def track_task(
        self,
        task: asyncio.Task[Any],
        event: asyncio.Event,
    ) -> None:
        self.idle_event.clear()
        self.pending_tasks.add(task)
        part = functools.partial(self._on_task_done, event=event)
        task.add_done_callback(part)

    def _on_task_done(
        self,
        task: asyncio.Task[Any],
        event: asyncio.Event,
    ) -> None:
        self.pending_tasks.discard(task)
        self._check_state()
        event.set()

    def _check_state(self) -> None:
        if not (self.pending_jobs or self.pending_tasks):
            self.idle_event.set()



import asyncio
from collections.abc import Callable, Mapping
from contextlib import AbstractAsyncContextManager
from typing import Any, TypeAlias, TypeVar

AppType = TypeVar("AppType")

StatelessLifespan: TypeAlias = Callable[
    [AppType],
    AbstractAsyncContextManager[None],
]
StatefulLifespan: TypeAlias = Callable[
    [AppType],
    AbstractAsyncContextManager[Mapping[str, Any]],
]
Lifespan: TypeAlias = StatelessLifespan[AppType] | StatefulLifespan[AppType]
LoopFactory: TypeAlias = Callable[[], asyncio.AbstractEventLoop]



from __future__ import annotations

import multiprocessing
import sys
from concurrent.futures import ProcessPoolExecutor, ThreadPoolExecutor
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any, TypedDict

from typing_extensions import NotRequired

from jobify._internal.common.constants import INFINITY
from jobify._internal.scheduler.misfire_policy import (
    GracePolicy,
    MisfirePolicy,
)

if TYPE_CHECKING:
    from collections.abc import Mapping
    from datetime import datetime
    from zoneinfo import ZoneInfo

    from jobify._internal.common.constants import RunMode
    from jobify._internal.common.types import LoopFactory
    from jobify._internal.cron_parser import CronFactory
    from jobify._internal.serializers.base import Serializer
    from jobify._internal.storage.abc import Storage
    from jobify._internal.typeadapter.base import Dumper, Loader


@dataclass(slots=True, kw_only=True)
class WorkerPools:
    _processpool: ProcessPoolExecutor | None
    threadpool: ThreadPoolExecutor | None = None

    @property
    def processpool(self) -> ProcessPoolExecutor:  # pragma: no cover
        if self._processpool is None:
            if sys.platform in ("win32", "darwin"):
                start_method = "spawn"
            elif "forkserver" in multiprocessing.get_all_start_methods():
                start_method = "forkserver"
            else:
                start_method = "spawn"
            mp_ctx = multiprocessing.get_context(start_method)
            self._processpool = ProcessPoolExecutor(mp_context=mp_ctx)
        return self._processpool

    def close(self) -> None:
        if self._processpool is not None:
            self._processpool.shutdown(wait=True, cancel_futures=True)
            self._processpool = None


@dataclass(slots=True, kw_only=True)
class JobifyConfiguration:
    tz: ZoneInfo
    dumper: Dumper
    loader: Loader
    storage: Storage
    getloop: LoopFactory
    serializer: Serializer
    worker_pools: WorkerPools
    cron_factory: CronFactory
    app_started: bool = False


@dataclass(slots=True, kw_only=True)
class Cron:
    expression: str = field(kw_only=False)
    max_runs: int = INFINITY
    max_failures: int = 10
    misfire_policy: MisfirePolicy | GracePolicy = MisfirePolicy.ONCE
    start_date: datetime | None = None

    def __post_init__(self) -> None:
        if self.max_failures < 1:
            msg = (
                "max_cron_failures must be >= 1."
                " Use 1 for 'stop on first error'."
            )
            raise ValueError(msg)


class RouteOptions(TypedDict):
    name: NotRequired[str]
    cron: NotRequired[Cron | str]
    retry: NotRequired[int]
    timeout: NotRequired[float]
    durable: NotRequired[bool]
    run_mode: NotRequired[RunMode]
    metadata: NotRequired[Mapping[str, Any]]
